/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.common.binaryfile;

import static de.ozgcloud.common.binaryfile.GrpcBinaryFileServerDownloader.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Random;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.core.task.TaskExecutor;

import com.google.protobuf.ByteString;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.common.test.ReflectionTestUtils;
import io.grpc.stub.CallStreamObserver;
import lombok.SneakyThrows;

class GrpcBinaryFileServerDownloaderTest {

	@Mock
	private CallStreamObserver<GrpcResponseDummy> callObserver;
	@Mock
	private Function<ByteString, GrpcResponseDummy> chunkBuilder;
	@Mock
	private Consumer<OutputStream> downloadConsumer;
	@Mock
	private TaskExecutor taskExecutor;

	private GrpcBinaryFileServerDownloader<GrpcResponseDummy> downloader;

	@BeforeEach
	void init() {
		downloader = spy(GrpcBinaryFileServerDownloader.<GrpcResponseDummy>builder().callObserver(callObserver).downloadConsumer(downloadConsumer)
				.chunkBuilder(chunkBuilder).taskExecutor(taskExecutor).build());
	}

	@Nested
	class TestStart {

		@BeforeEach
		void init() {
			doNothing().when(downloader).doStart();
		}

		@Test
		void shouldCallDoStart() {

			downloader.start();

			verify(downloader).doStart();
		}

		@Test
		void shouldNotStartWhenStarted() {
			downloader.start();
			downloader.start();

			verify(downloader, times(1)).doStart();
		}
	}

	@Nested
	class TestIsNotStarted {

		@Test
		void shouldReturnTrueIfNotStarted() {
			var result = downloader.isNotStarted();

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseIfStarted() {
			downloader.isNotStarted();

			var result = downloader.isNotStarted();

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestDoStart {

		@Captor
		private ArgumentCaptor<Runnable> runnableCaptor;

		@BeforeEach
		void init() {
			doNothing().when(downloader).safelySetupStreams();
		}

		@SneakyThrows
		@Test
		void shouldSafelySetupStreams() {
			downloader.doStart();

			verify(downloader).safelySetupStreams();
		}

		@Test
		void shouldCallStartDownload() {
			doNothing().when(downloader).startDownload();

			downloader.doStart();

			verify(taskExecutor).execute(runnableCaptor.capture());
			runnableCaptor.getValue().run();
			verify(downloader).startDownload();
		}

		@SneakyThrows
		@Test
		void shouldSetOnReadyHandler() {
			downloader.doStart();

			verify(callObserver).setOnReadyHandler(runnableCaptor.capture());
			assertThat(runnableCaptor.getValue()).isNotNull();
		}
	}

	@Nested
	class TestSafelySetupStreams {

		@SneakyThrows
		@Test
		void shouldSetupStreams() {
			doNothing().when(downloader).setupStreams();

			safelySetupStreams();

			verify(downloader).setupStreams();
		}

		@Nested
		class OnException {

			@Mock
			private PipedOutputStream outputStream;
			@Mock
			private PipedInputStream inputStream;

			private final IOException exception = new IOException();

			@SneakyThrows
			@BeforeEach
			void init() {
				setInputStreamField(inputStream);
				setOutputStreamField(outputStream);
				doThrow(exception).when(downloader).setupStreams();
			}

			@Test
			void shouldThrowBinaryFileDownloadException() {
				assertThatThrownBy(() -> downloader.safelySetupStreams()).isInstanceOf(TechnicalException.class).hasCause(exception);
			}

			@SneakyThrows
			@Test
			void shouldCloseOutputStream() {
				catchThrowable(TestSafelySetupStreams.this::safelySetupStreams);

				verify(outputStream).close();
			}

			@SneakyThrows
			@Test
			void shouldCloseInputStream() {
				catchThrowable(TestSafelySetupStreams.this::safelySetupStreams);

				verify(inputStream).close();
			}
		}

		private void safelySetupStreams() {
			downloader.safelySetupStreams();
		}
	}

	@Nested
	class TestStartDownload {

		@Mock
		private PipedOutputStream outputStream;

		@BeforeEach
		void init() {
			setOutputStreamField(outputStream);
		}

		@Test
		void shouldErrorBeInitiallyNull() {
			assertThat(getDownloadError()).isNull();
		}

		@Nested
		class OnNoException {

			@BeforeEach
			void init() {
				doNothing().when(downloader).doDownload();
			}

			@Test
			void shouldDoDownload() {
				downloader.startDownload();

				verify(downloader).doDownload();
			}

			@SneakyThrows
			@Test
			void shouldCloseOutputStream() {
				downloader.startDownload();

				verify(outputStream).close();
			}
		}

		@Nested
		class OnException {

			private final TechnicalException exception = new TechnicalException("error");

			@BeforeEach
			void init() {
				doThrow(exception).when(downloader).doDownload();
			}

			@Test
			void shouldSetError() {
				downloader.startDownload();

				assertThat(getDownloadError()).isInstanceOf(TechnicalException.class).hasCause(exception);
			}

			@SneakyThrows
			@Test
			void shouldCloseOutputStream() {
				downloader.startDownload();

				verify(outputStream).close();
			}
		}
	}

	@Nested
	class TestDoDownload {

		@Mock
		private PipedOutputStream outputStream;

		@BeforeEach
		void mock() {
			setOutputStreamField(outputStream);
		}

		@Test
		void shouldCallDownloadConsumer() {
			downloader.doDownload();

			verify(downloadConsumer).accept(outputStream);
		}
	}

	@Nested
	class TestSendChunks {

		@Nested
		class OnNoException {

			@SneakyThrows
			@BeforeEach
			void init() {
				doNothing().when(downloader).doSendChunks();
			}

			@SneakyThrows
			@Test
			void shouldDoSendChunks() {
				downloader.sendChunks();

				verify(downloader).doSendChunks();
			}
		}

		@Nested
		class OnException {

			private final IOException exception = new IOException();
			@Captor
			private ArgumentCaptor<TechnicalException> argumentCaptor;

			@SneakyThrows
			@BeforeEach
			void init() {
				doThrow(exception).when(downloader).doSendChunks();
				doNothing().when(downloader).handleError(any());
			}

			@Test
			void shouldHandleError() {
				downloader.sendChunks();

				verify(downloader).handleError(argumentCaptor.capture());
				assertThat(argumentCaptor.getValue()).isInstanceOf(TechnicalException.class).hasCause(exception);
			}
		}
	}

	@Nested
	class TestDoSendChunks {

		@SneakyThrows
		@Test
		void shouldCheckIfCanSendChunks() {
			doReturn(false).when(downloader).canSendChunks();

			downloader.doSendChunks();

			verify(downloader).canSendChunks();
		}

		@SneakyThrows
		@Test
		void shouldNotProcessIfCannotSendChunks() {
			doReturn(false).when(downloader).canSendChunks();

			downloader.doSendChunks();

			verify(downloader, never()).processDataFromInputStream();
		}

		@SneakyThrows
		@Test
		void shouldProcessAsLongAsCanSendChunks() {
			doReturn(true, true, false).when(downloader).canSendChunks();
			doNothing().when(downloader).processDataFromInputStream();

			downloader.doSendChunks();

			verify(downloader, times(2)).processDataFromInputStream();
		}
	}

	@Nested
	class TestProcessDataFromInputStream {

		@Mock
		private PipedInputStream inputStream;

		@Nested
		class OnEndOfStreamReached {

			@SneakyThrows
			@BeforeEach
			void init() {
				doNothing().when(downloader).finishProcessing();
				when(inputStream.read(any())).thenReturn(END_OF_STREAM);
				setInputStreamField(inputStream);
			}

			@SneakyThrows
			@Test
			void shouldFinishProcessing() {
				downloader.processDataFromInputStream();

				verify(downloader).finishProcessing();
			}

			@SneakyThrows
			@Test
			void shouldNotSendBytesToCallObserver() {
				downloader.processDataFromInputStream();

				verify(downloader, never()).sendBytesToCallObserver(anyInt());
			}
		}

		@Nested
		class OnNoBytesWereReceived {

			@SneakyThrows
			@BeforeEach
			void init() {
				when(inputStream.read(any())).thenReturn(NOTHING_READ);
				setInputStreamField(inputStream);
			}

			@SneakyThrows
			@Test
			void shouldNotSendBytesToCallObserver() {
				downloader.processDataFromInputStream();

				verify(downloader, never()).sendBytesToCallObserver(anyInt());
			}
		}

		@Nested
		class OnBytesWereReceived {

			private final int bytesRead = 20;

			@SneakyThrows
			@BeforeEach
			void mock() {
				when(inputStream.read(any())).thenReturn(bytesRead);
				setInputStreamField(inputStream);
			}

			@SneakyThrows
			@Test
			void shouldSendBytesToCallObserver() {
				downloader.processDataFromInputStream();

				verify(downloader).sendBytesToCallObserver(bytesRead);
			}
		}
	}

	@Nested
	class TestSendBytesToCallObserver {

		@Captor
		private ArgumentCaptor<ByteString> byteStringCaptor;
		private final int bytesRead = 20;
		private final byte[] buffer = new byte[bytesRead];
		private final GrpcResponseDummy grpcResponseDummy = new GrpcResponseDummy();

		@BeforeEach
		void init() {
			new Random().nextBytes(buffer);
			ReflectionTestUtils.setField(downloader, "buffer", buffer);
		}

		@SneakyThrows
		@Test
		void shouldCallChunkBuilder() {
			downloader.sendBytesToCallObserver(bytesRead);

			verify(chunkBuilder).apply(byteStringCaptor.capture());
			assertThat(byteStringCaptor.getValue().toByteArray()).isEqualTo(buffer);
		}

		@SneakyThrows
		@Test
		void shouldCallOnNext() {
			when(chunkBuilder.apply(any())).thenReturn(grpcResponseDummy);

			downloader.sendBytesToCallObserver(bytesRead);

			verify(callObserver).onNext(grpcResponseDummy);
		}
	}

	@Nested
	class TestCanSendChunks {

		@ParameterizedTest
		@MethodSource("provideArguments")
		void shouldReturnValue(boolean requestFinished, boolean ready, boolean expected) {
			setRequestFinishedField(requestFinished);
			lenient().when(callObserver.isReady()).thenReturn(ready);

			var canSendChunks = downloader.canSendChunks();

			assertThat(canSendChunks).isEqualTo(expected);
		}

		private static Stream<Arguments> provideArguments() {
			return Stream.of(
				Arguments.of(false,  false, false),
				Arguments.of(false, true, true),
				Arguments.of(true, false, false),
				Arguments.of(true, true, false)
			);
		}
	}

	@Nested
	class TestFinishProcessing {

		@Nested
		class OnError {

			private final TechnicalException exception = new TechnicalException("error");

			@BeforeEach
			void init() {
				setDownloadErrorField(exception);
			}

			@Test
			void shouldThrowException() {
				assertThatThrownBy(downloader::finishProcessing).isSameAs(exception);
			}
		}

		@Nested
		class OnNoError {

			@BeforeEach
			void init() {
				doNothing().when(downloader).closeInputStream();
			}

			@Test
			void shouldNotCompleteRequestWithError() {
				downloader.finishProcessing();

				verify(downloader, never()).handleError(any());
			}

			@Test
			void shouldSetRequestFinished() {
				assertThat(getRequestFinished()).isFalse();

				downloader.finishProcessing();

				assertThat(getRequestFinished()).isTrue();
			}

			@Test
			void shouldCloseInputStream() {
				downloader.finishProcessing();

				verify(downloader).closeInputStream();
			}

			@Test
			void shouldNotifyObserver() {
				downloader.finishProcessing();

				verify(callObserver).onCompleted();
			}
		}
	}

	@Nested
	class TestHandleError {

		private final TechnicalException error = new TechnicalException("error");

		@BeforeEach
		void init() {
			doNothing().when(downloader).closeInputStream();
		}

		@Test
		void shouldSetRequestFinished() {
			assertThat(getRequestFinished()).isFalse();

			catchException(() -> downloader.handleError(error));

			assertThat(getRequestFinished()).isTrue();
		}

		@Test
		void shouldCloseInputStream() {
			catchException(() -> downloader.handleError(error));

			verify(downloader).closeInputStream();
		}

		@Test
		void shouldThrowException() {
			assertThatThrownBy(() -> downloader.handleError(error)).isSameAs(error);
		}
	}

	private void setOutputStreamField(OutputStream outputStream) {
		ReflectionTestUtils.setField(downloader, "outputStream", outputStream);
	}

	private void setInputStreamField(InputStream inputStream) {
		ReflectionTestUtils.setField(downloader, "inputStream", inputStream);
	}

	private void setRequestFinishedField(boolean requestFinished) {
		ReflectionTestUtils.setField(downloader, "requestFinished", new AtomicBoolean(requestFinished));
	}

	private boolean getRequestFinished() {
		return ReflectionTestUtils.getField(downloader, "requestFinished", AtomicBoolean.class).get();
	}

	private void setDownloadErrorField(TechnicalException error) {
		ReflectionTestUtils.setField(downloader, "downloadError", new AtomicReference<>(error));
	}

	private TechnicalException getDownloadError() {
		return (TechnicalException) ReflectionTestUtils.getField(downloader, "downloadError", AtomicReference.class).get();
	}

	private static class GrpcResponseDummy {
	}
}