/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.common.pdf.fop;

import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Objects;

import javax.xml.XMLConstants;
import jakarta.xml.bind.JAXBContext;
import jakarta.xml.bind.JAXBException;
import jakarta.xml.bind.util.JAXBSource;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.URIResolver;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.fop.apps.FOPException;
import org.apache.fop.apps.FOUserAgent;
import org.apache.fop.apps.Fop;
import org.apache.fop.apps.FopFactory;
import org.apache.fop.apps.FopFactoryBuilder;
import org.apache.fop.configuration.Configuration;
import org.apache.fop.configuration.ConfigurationException;
import org.apache.fop.configuration.DefaultConfigurationBuilder;
import org.apache.xmlgraphics.util.MimeConstants;
import org.springframework.lang.NonNull;
import org.springframework.stereotype.Service;

import de.ozgcloud.common.errorhandling.TechnicalException;

import jakarta.enterprise.context.ApplicationScoped;
import lombok.extern.log4j.Log4j2;

@Service
@ApplicationScoped
@Log4j2
public class FopPdfService {

	private static final String OUTPUT_FORMAT = MimeConstants.MIME_PDF;
	private static final String BASE_URI = "classpath://fop";
	private static final String MASTER_TEMPLATE = "fop/master-template.xsl";
	private static final String CONFIG_FILE = "fop/fop_conf.xml";
	private static final String TRANSFORMER_FACTORY_CLASS = "com.sun.org.apache.xalan.internal.xsltc.trax.TransformerFactoryImpl";

	private static final FopFactory fopFactory = createFactory();

	public OutputStream createPdf(@NonNull InputStream template, @NonNull OutputStream out, @NonNull Object model) {
		var bufferedOut = new BufferedOutputStream(out);
		renderPdf(template, bufferedOut, model);
		return flushStream(bufferedOut);
	}

	private BufferedOutputStream flushStream(BufferedOutputStream stream) {
		try {
			stream.flush();
		} catch (IOException e) {
			LOG.warn("Error on flushing stream", e);
		}
		return stream;
	}

	void renderPdf(InputStream template, OutputStream out, Object model) {
		try {
			createTransformer(template).transform(createJaxbSource(model), createSaxResult(out));
		} catch (TransformerException e) {
			throw new TechnicalException("Error on transforming to PDF.", e);
		}
	}

	Source createJaxbSource(Object model) {
		try {
			return new JAXBSource(JAXBContext.newInstance(model.getClass()), model);
		} catch (JAXBException e) {
			throw new TechnicalException("Error creating jaxb Source", e);
		}
	}

	Result createSaxResult(OutputStream out) {
		try {
			Fop fop = fopFactory.newFop(OUTPUT_FORMAT, createUserAgent(), out);
			return new SAXResult(fop.getDefaultHandler());
		} catch (FOPException e) {
			throw new TechnicalException("Error creating Result-Object", e);
		}
	}

	FOUserAgent createUserAgent() {
		var userAgent = fopFactory.newFOUserAgent();

		userAgent.setTitle("-");
		return userAgent;
	}

	static FopFactory createFactory() {
		try {
			return new FopFactoryBuilder(new URI(BASE_URI))
					.setConfiguration(createConfiguration())//
					.build();
		} catch (URISyntaxException e) {
			throw new TechnicalException("Error on creating Fop-Factory.", e);
		}
	}

	static Configuration createConfiguration() {
		try (InputStream confStream = getConfigAsStream()) {
			return new DefaultConfigurationBuilder().build(confStream);
		} catch (ConfigurationException | IOException e) {
			throw new TechnicalException("Error on building configuration", e);
		}
	}

	private static InputStream getConfigAsStream() {
		InputStream confStream = FopPdfService.class.getClassLoader().getResourceAsStream(CONFIG_FILE);
		if (Objects.isNull(confStream)) {
			throw new TechnicalException("Cannot find fop config on " + CONFIG_FILE);
		}
		return confStream;
	}

	Transformer createTransformer(InputStream template) {
		try {
			var factory = TransformerFactory.newInstance(TRANSFORMER_FACTORY_CLASS, FopPdfService.class.getClassLoader());
			factory.setAttribute(XMLConstants.ACCESS_EXTERNAL_DTD, "");
			factory.setAttribute(XMLConstants.ACCESS_EXTERNAL_STYLESHEET, "");
			factory.setURIResolver(createResourceResolver(template));
			return factory.newTransformer(createMasterSource());
		} catch (TransformerConfigurationException | TransformerFactoryConfigurationError e) {
			throw new TechnicalException("Cannot create transformer.", e);
		}
	}

	static StreamSource createMasterSource() {
		InputStream templateStream = FopPdfService.class.getClassLoader().getResourceAsStream(MASTER_TEMPLATE);
		if (Objects.isNull(templateStream)) {
			throw new TechnicalException("Cannot find master template on " + MASTER_TEMPLATE);
		}

		return new StreamSource(templateStream);
	}

	static URIResolver createResourceResolver(InputStream template) {
		return new URIResolver() {

			@Override
			public Source resolve(String href, String base) throws TransformerException {
				return new StreamSource(template);
			}

		};
	}
}
